//PROB3.CPP  My solution to the Part 3 Exercise

#include <iostream.h>
#include <string.h>

const int maxStudents = 20;  //max no. of students in one class
const int maxCourses = 6;    //max no. of courses student can take

class Student;                         //Note 1

class Course
{
   friend class Student;
  public:
   Course()                            //Note 2
   {
      noStudents = 0;
      semesterHours = 3.0;   //default is 3.0
   }
  ~Course()
   {
      //nothing to do here either
   }

   //hours - return or set the number of hours in this class
   int hours()                         //Note 3
   {
      return semesterHours;
   }
   int hours(int newValue)
   {
      int oldValue = semesterHours;    //it's customary to...
      semesterHours = newValue;
      return oldValue;                 //...return old value
   }

   //grade - return the average for the class or the grade of a
   //        single student or set the grade of a single student
   float grade();                      //Note 4
   float grade(Student &s);
   void  grade(Student &s, float grade);

  protected:
   //member functions
   //add - add a student to the role
   int add(Student &s);

   //findStudent - find student in role
   int findStudent(Student &s);

   //data members
   Student *pStudents[maxStudents];
   float    classGrade[maxStudents];
   int      semesterHours;
   int      noStudents;
};

class Student
{
  public:
   Student()
   {
      //just zero out everything
      noCourses = 0;
      sName[0] = '\0';
   }
  ~Student()
   {
      //nothing to do
   }

   //return the grade in a given course or average in all courses
   float grade(Course &c)              //Note 5
   {
      return c.grade(*this);
   }
   float grade();

   //return the grade compared to the class average
   float curvedGrade(Course &c);

   //add a student to a course
   void add(Course &c);

   //name - read and set name
   char *name()
   {
      return sName;
   }
   void  name(char *pName);

  protected:
   char    sName[20];
   Course *pClasses[maxCourses];
   int     noCourses;
};

//--------------Course member functions---------------------
//grade - return the average for the class
float Course::grade()
{
   //if there are no students, forget it now
   if (noStudents == 0)
   {
      return 0.0;
   }

   //add 'em all up and return the average
   float accumGrade = 0.0;
   for (int i = 0; i < noStudents; i++)
   {
      accumGrade += classGrade[i];
   }
   return accumGrade / noStudents;
}

//grade(Student*) - return the grade of a single student
float Course::grade(Student &s)
{
   //first look him up in the student role
   int offset = findStudent(s);

   //return his grade or 0 if he wasn't found
   return (offset >= 0) ? classGrade[offset] : 0.0;
}

//grade(Student, grade) - set the grade of a single student
void Course::grade(Student &s, float grade)
{
   int offset = findStudent(s);
   if (offset >= 0)
   {
      classGrade[offset] = grade;
   }
}

//add - add a student to the class role; return 1 if
//      it works and a zero otherwise
int Course::add(Student &s)
{
   if (noStudents >= maxStudents)
   {
      cout << "No more room in class...maybe next semester\n";
      return 0;
   }
   pStudents[noStudents++] = &s;
   return 1;
}

//findStudent - find a student in the look up table;
//              return offset or -1 if can't find
int Course::findStudent(Student &s)
{
   for (int i = 0; i < noStudents; i++)
   {
      if (pStudents[i] == &s)
      {
         return i;
      }
   }
   return -1;
}

//--------------Student member functions-------------------
//grade() - return the average grade for all courses
float Student::grade()
{
   int hours;
   int noHours = 0;
   float accumGrade = 0.0;
   for (int i = 0; i < noCourses; i++)
   {
      hours = pClasses[i]->hours();
      accumGrade += pClasses[i]->grade(*this) * hours;
      noHours += hours;
   }
   return (noHours) ? (accumGrade / noHours): 0.0;
}

//add - add a student to a class
void Student::add(Course &c)
{
   if (noCourses >= maxCourses)
   {
      cout << "no more room in the inn tonite Joseph\n";
      return;
   }
   if (c.add(*this))
   {
      pClasses[noCourses++] = &c;
   }
}

//curvedGrade - return the grade curved by the average
float Student::curvedGrade(Course &c)
{
   //set the average to 3.0
   float g = 3.0 * grade(c) / c.grade();

   //don't let it go over 4.0
   if (g > 4.0)
   {
      g = 4.0;
   }
   return g;
}

//name(char*) - store off the name provided
void  Student::name(char *pName)
{
   strncpy(sName, pName, sizeof(sName) - 1);
   sName[sizeof(sName) - 1] = '\0';
}

//--------------------test program-------------------------
int main()
{
   //declare a course
   Course geo101;

   //give it a couple of students
   Student harry, anne;
   harry.name("Harry");
   harry.add(geo101);
   anne.name("Anne");
   anne.add(geo101);

   //grade them
   geo101.grade(harry, 3.0);
   geo101.grade(anne, 2.5);

   //now let's look at the averages
   cout<<"average geo101 grade = "<<geo101.grade()          <<"\n";
   cout<<"anne's grade         = "<<anne.grade(geo101)      <<"\n";
   cout<<"anne's curved grade  = "<<anne.curvedGrade(geo101)<<"\n";

   //let's look at harry's other course
   Course engl201;
   engl201.hours(4.0);
   harry.add(engl201);
   engl201.grade(harry, 2.0);
   cout << "harry's average grade = " << harry.grade() << "\n";

   return 0;
}